(function($, PHD, window, undefined) {
    
    /*
    *   An object that triggers namespaced events on an interval
    */

    var STOPPED = 0,
        RUNNING = 1,
        PAUSED = 2,
        MANUAL = 3;

    var defaults = {
        namespace: "phdtimer",
        interval: 2000,
        minimumInterval: 1000,
        pauseOnWizardOpen: true, // Pause the timer when a wizard is opened -- resume when it is closed
        stopOnWizardOpen: false, // Stop the timer completely when a wizard is opened
        pauseOnBroadcast: true, // Wait for the subscriber to signal that it is done
        stopOnAjaxAbort: true, // Stop when an ajaxaborted or ajaxtimeout event is triggered
        autostart: false,
        refreshControl: null,
        beforeBroadcast: function() { return true; },
        afterBroadcast: function() { return true; }
    };

    var Timer = function Timer(options) {
        if(this instanceof Timer) {
            options = options || {};

            this.options = $.extend({}, defaults, options);

	    console.log('=====>Timer for: ' + this.options.namespace);

            this._interval = isInt(this.options.interval) ? parseInt(this.options.interval, 10) : defaults.interval;
            this._status = STOPPED;
            this._timeout = null;
            this._promise = null;
            this._ready = false;
        } else {
            return new Timer(options).init();
        }
    };

    Timer.prototype = {
        init: function() {
            var self = this;

            if(this.options.refreshControl) {
                this.$refreshControl = this.options.refreshControl;
            } else {
                this.$refreshControl = $("<i>", {
                    "class": "icon-refreshalt timer-refresh"
                });
            }

            if(this.options.pauseOnWizardOpen) {
                $(document).on("wizardopen", $.proxy(this.pause, this));
                $(document).on("wizardclose", $.proxy(this.resume, this));
            } else if(this.options.stopOnWizardOpen) {
                $(document).on("wizardopen", $.proxy(this.stop, this));
            }

            if(this.options.stopOnAjaxAbort) {
                $(document).on("ajaxaborted ajaxtimeout", $.proxy(this.stop, this));
            }

            if(this._interval < this.options.minimumInterval) {
                this.setManual();
            } else if(this.options.autostart) {
                this.start();
            }

            return this;
        },

        start: function() {
            if(this._status === STOPPED) {
                this._status = RUNNING;
                this._setTimer();
                $(document).trigger("timerstart." + this.options.namespace, [this, this._interval]);
            }
            return this;
        },

        stop: function() {
            if(this._timeout !== null && this._status !== STOPPED) {
                this._status = STOPPED;
                clearTimeout(this._timeout);
                this._timeout = null;
                if(this._promise !== null) {
                    this._promise.reject();
                    this._promise = null;
                }
                $(document).trigger("timerstop." + this.options.namespace, [this, this._interval]);
            }
            return this;
        },

        pause: function() {
            if(this._timeout !== null && this._status === RUNNING) {
                this._status = PAUSED;
                this._promise = $.Deferred();
                $(document).trigger("timerpause." + this.options.namespace, [this, this._interval]);
            }
            return this;
        },

        resume: function() {
            if(this._promise !== null && this._status === PAUSED) {
                var resolved = (this._promise.state() === "resolved");
                this._status = RUNNING;
                $(document).trigger("timerresume." + this.options.namespace, [this, this._interval]);
                this._promise = null;
                if(resolved) {
                    this._broadcast();
                }
                return this;
            }
        },

        reset: function() {
            this.stop();
            this.start();
        },

        setInterval: function(interval) {
            if(isInt(interval)) {
                interval = parseInt(interval, 10);
                if(interval < this.options.minimumInterval) {
                    this.options.interval = this._interval = 0;
                    this.setManual();
                } else {
                    this.options.interval = this._interval = interval;
                    if(this._status === MANUAL) {
                        this.unsetManual();
                    }
                }
            } else {
                throw new TypeError("Timer interval must be an integer");
            }
            return this;
        },

        setManual: function() {
            if(this._status !== MANUAL) {
                this.stop();
                this._status = MANUAL;
                this.$refreshControl
                    .on("click." + this.options.namespace, $.proxy(this._broadcast, this));
                $(document).trigger("timersetmanual." + this.options.namespace, [this.$refreshControl, this]);
            }
            return this;
        },

        unsetManual: function() {
            if(this._status === MANUAL) {
                this.$refreshControl
                    .off("click." + this.options.namespace);
                this._status = STOPPED;
                this.start();
                $(document).trigger("timerunsetmanual." + this.options.namespace, [this.$refreshControl, this]);
                return this;
            }
            return this;
        },

        getInterval: function(interval) {
            return this._interval;            
        },

        getStatus: function() {
            return this._status;
        },

        _tick: function() {
            if(this._status === RUNNING && this._promise === null) {
                this._broadcast();
            } else if(this._promise !== null) {
                this._promise.resolve();
            } else {
                this.stop();
            }
            return this;
        },

        _broadcast: function() {
            if(this._status === RUNNING) {
                if(this.options.beforeBroadcast()) {
                    this._timeout = null;
                    this._setTimer();
                    if(this.options.pauseOnBroadcast) {
                        this.pause();
                    }
                    $(document).trigger("timertick." + this.options.namespace, [this, this._interval]);
                    this.options.afterBroadcast();
                }
            } else if(this._status === MANUAL) {
                if(this.options.beforeBroadcast()) {
                    $(document).trigger("timertick." + this.options.namespace, [this, this._interval]);
                    this.options.afterBroadcast();
                }
            }
        },

        _setTimer: function() {
            if(this._interval < this.options.minimumInterval) {
                throw new Error("Interval too small");
            }
            this._timeout = this._doTimeout(this._tick, this._interval);
        },

        _doTimeout: function(cb, interval) {
            var self = this,
                _args = Array.prototype.slice.call(arguments, 2);

            return setTimeout(function() { cb.apply(self, _args); }, interval);
        },

	// lmc
	rerun: function(init, change) {
	    if (this._status === PAUSED) {
		this.resume();
	    } else if (this._status === STOPPED) {
		this.start();
	    } else {
		this.reset();
	    }
	}

    };

    PHD.Timer = Timer;

    var isInt = function(n) {
        return !isNaN(parseInt(n, 10));
    };

})(jQuery, window.PHD || (window.PHD = {}), window);
